@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 1996-2000, 2003, 2004, 2005, 2006, 2007 by                    */
+/*  Copyright 1996-2000, 2003-2007, 2009-2012 by                            */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -33,23 +33,23 @@
 #ifdef CEIL
 #undef CEIL
 #endif
-#define CEIL( x )   ( ( (x) + 63 ) >> 6 )
+#define CEIL( x )  ( ( (x) + 63 ) >> 6 )
 
-#define INIT_SIZE( size, start_x, start_y, step_x, step_y, x, y )        \
-          do {                                                           \
-            start_x = 4;                                                 \
-            start_y = CEIL( size->metrics.height ) + 2 * HEADER_HEIGHT;  \
-            step_x  = CEIL( size->metrics.max_advance );                 \
-            step_y  = CEIL( size->metrics.height ) + 4;                  \
-                                                                         \
-            x = start_x;                                                 \
-            y = start_y;                                                 \
+#define INIT_SIZE( size, start_x, start_y, step_y, x, y )               \
+          do {                                                          \
+            start_x = 4;                                                \
+            start_y = CEIL( size->metrics.height ) + 3 * HEADER_HEIGHT; \
+            step_y  = CEIL( size->metrics.height ) + 4;                 \
+                                                                        \
+            x = start_x;                                                \
+            y = start_y;                                                \
           } while ( 0 )
 
-#define X_TOO_LONG( x, size, display) \
-          ( ( x ) + ( ( size )->metrics.max_advance >> 6 ) > ( display )->bitmap->width )
-#define Y_TOO_LONG( y, size, display) \
-          ( ( y ) >= ( display )->bitmap->rows )
+#define X_TOO_LONG( x, size, display )                   \
+          ( (x) + ( (size)->metrics.max_advance >> 6 ) > \
+            (display)->bitmap->width )
+#define Y_TOO_LONG( y, size, display )       \
+          ( (y) >= (display)->bitmap->rows )
 
 #ifdef _WIN32
 #define snprintf  _snprintf
@@ -60,6 +60,7 @@
   {
     RENDER_MODE_ALL = 0,
     RENDER_MODE_EMBOLDEN,
+    RENDER_MODE_SLANTED,
     RENDER_MODE_STROKE,
     RENDER_MODE_TEXT,
     RENDER_MODE_WATERFALL,
@@ -68,24 +69,34 @@
 
   static struct  status_
   {
-    int          render_mode;
-    FT_Encoding  encoding;
-    int          res;
-    int          ptsize;            /* current point size, 26.6 format */
-    int          lcd_mode;
-    double       gamma;
+    int            render_mode;
+    FT_Encoding    encoding;
+    int            res;
+    int            ptsize;            /* current point size, 26.6 format */
+    int            lcd_mode;
+    double         gamma;
+    double         xbold_factor;
+    double         ybold_factor;
+    double         slant;
 
-    int          debug;
-    int          trace_level;
-    int          font_index;
-    int          dump_cache_stats;  /* do we need to dump cache statistics? */
-    int          Num;               /* current first index */
-    char*        header;
-    char         header_buffer[256];
-    int          Fail;
-    int          preload;
+    int            debug;
+    int            trace_level;
+    int            font_index;
+    int            dump_cache_stats;  /* do we need to dump cache statistics? */
+    int            Num;               /* current first index */
+    char*          header;
+    char           header_buffer[256];
+    int            Fail;
+    int            preload;
 
-  } status = { RENDER_MODE_ALL, FT_ENCODING_NONE, 72, 48, -1, 1.0, 0, 0, 0, 0, 0, NULL, { 0 }, 0, 0 };
+    int            use_custom_lcd_filter;
+    unsigned char  filter_weights[5];
+    int            fw_index;
+
+  } status = { RENDER_MODE_ALL, FT_ENCODING_NONE, 72, 48, -1,
+               1.0, 0.04, 0.04, 0.22,
+               0, 0, 0, 0, 0, NULL, { 0 }, 0, 0,
+               0, { 0x10, 0x40, 0x70, 0x40, 0x10 }, 2 };
 
 
   static FTDemo_Display*  display;
@@ -93,15 +104,15 @@
 
 
   static const unsigned char*  Text = (unsigned char*)
-    "The quick brown fox jumps over the lazy dog 0123456789 "
-    "\342\352\356\373\364\344\353\357\366\374\377\340\371\351\350\347 "
-    "&#~\"\'(-`_^@)=+\260 ABCDEFGHIJKLMNOPQRSTUVWXYZ "
-    "$\243^\250*\265\371%!\247:/;.,?<>";
+    "The quick brown fox jumps over the lazy dog 0123456789"
+    " \342\352\356\373\364\344\353\357\366\374\377\340\371\351\350\347"
+    " &#~\"\'(-`_^@)=+\260 ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+    " $\243^\250*\265\371%!\247:/;.,?<>";
 
 
 
   static void
-  Fatal( const char* message )
+  Fatal( const char*  message )
   {
     FTDemo_Display_Done( display );
     FTDemo_Done( handle );
@@ -113,7 +124,7 @@
   Render_Stroke( int  num_indices,
                  int  first_index )
   {
-    int         start_x, start_y, step_x, step_y, x, y;
+    int         start_x, start_y, step_y, x, y;
     int         i;
     FT_Size     size;
     FT_Stroker  stroker = NULL;
@@ -127,7 +138,7 @@
       return error;
     }
 
-    INIT_SIZE( size, start_x, start_y, step_x, step_y, x, y );
+    INIT_SIZE( size, start_x, start_y, step_y, x, y );
 
     i = first_index;
 
@@ -159,6 +170,7 @@
       {
         FT_Glyph  glyph;
 
+
         error = FT_Get_Glyph( slot, &glyph );
         if ( error )
           goto Next;
@@ -186,7 +198,7 @@
       }
       else
       {
-Next:
+    Next:
         status.Fail++;
       }
 
@@ -202,12 +214,12 @@ Next:
 
 
   static FT_Error
-  Render_Embolden( int  num_indices,
-                   int  first_index )
+  Render_Slanted( int  num_indices,
+                  int  first_index )
   {
-    int       start_x, start_y, step_x, step_y, x, y;
-    int       i;
-    FT_Size   size;
+    int      start_x, start_y, step_y, x, y;
+    int      i;
+    FT_Size  size;
 
 
     error = FTDemo_Get_Size( handle, &size );
@@ -218,14 +230,15 @@ Next:
       return error;
     }
 
-    INIT_SIZE( size, start_x, start_y, step_x, step_y, x, y );
+    INIT_SIZE( size, start_x, start_y, step_y, x, y );
 
     i = first_index;
 
     while ( i < num_indices )
     {
-      int           gindex;
-      FT_Face       face = size->face;
+      int      gindex;
+      FT_Face  face = size->face;
+
 
       if ( handle->encoding == FT_ENCODING_NONE )
         gindex = i;
@@ -235,7 +248,32 @@ Next:
       error = FT_Load_Glyph( face, gindex, handle->load_flags );
       if ( !error )
       {
-        FT_GlyphSlot_Embolden( face->glyph );
+        FT_Matrix    shear;
+        FT_Outline*  outline;
+
+
+        /***************************************************************/
+        /*                                                             */
+        /*  2*2 affine transformation matrix, 16.16 fixed float format */
+        /*                                                             */
+        /*  Shear matrix:                                              */
+        /*                                                             */
+        /*         | x' |     | 1  k |   | x |          x' = x + ky    */
+        /*         |    |  =  |      | * |   |   <==>                  */
+        /*         | y' |     | 0  1 |   | y |          y' = y         */
+        /*                                                             */
+        /*        outline'     shear    outline                        */
+        /*                                                             */
+        /***************************************************************/
+
+        shear.xx = 1 << 16;
+        shear.xy = (FT_Fixed)( status.slant * ( 1 << 16 ) );
+        shear.yx = 0;
+        shear.yy = 1 << 16;
+
+        outline = &(face->glyph)->outline;
+
+        FT_Outline_Transform( outline, &shear );
 
         error = FTDemo_Draw_Slot( handle, display, face->glyph, &x, &y );
 
@@ -261,12 +299,12 @@ Next:
 
 
   static FT_Error
-  Render_All( int  num_indices,
-              int  first_index )
+  Render_Embolden( int  num_indices,
+                   int  first_index )
   {
-    int         start_x, start_y, step_x, step_y, x, y;
-    int         i;
-    FT_Size     size;
+    int      start_x, start_y, step_y, x, y;
+    int      i;
+    FT_Size  size;
 
 
     error = FTDemo_Get_Size( handle, &size );
@@ -277,7 +315,117 @@ Next:
       return error;
     }
 
-    INIT_SIZE( size, start_x, start_y, step_x, step_y, x, y );
+    INIT_SIZE( size, start_x, start_y, step_y, x, y );
+
+    i = first_index;
+
+    while ( i < num_indices )
+    {
+      int      gindex;
+      FT_Face  face = size->face;
+
+
+      if ( handle->encoding == FT_ENCODING_NONE )
+        gindex = i;
+      else
+        gindex = FTDemo_Get_Index( handle, i );
+
+      error = FT_Load_Glyph( face, gindex, handle->load_flags );
+      if ( !error )
+      {
+        /* this is essentially the code of function */
+        /* `FT_GlyphSlot_Embolden'                  */
+
+        FT_GlyphSlot  slot    = face->glyph;
+        FT_Library    library = slot->library;
+        FT_Pos        xstr, ystr;
+
+
+        if ( slot->format != FT_GLYPH_FORMAT_OUTLINE &&
+             slot->format != FT_GLYPH_FORMAT_BITMAP )
+          goto Next;
+
+        ystr = FT_MulFix( face->units_per_EM,
+                          face->size->metrics.y_scale );
+        xstr = (FT_Fixed)( ystr * status.xbold_factor );
+        ystr = (FT_Fixed)( ystr * status.ybold_factor );
+
+        if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
+        {
+          error = FT_Outline_EmboldenXY( &slot->outline, xstr, ystr );
+          /* ignore error */
+        }
+        else if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
+        {
+          /* round to full pixels */
+          xstr &= ~63;
+          ystr &= ~63;
+
+          error = FT_GlyphSlot_Own_Bitmap( slot );
+          if ( error )
+            goto Next;
+
+          error = FT_Bitmap_Embolden( library, &slot->bitmap, xstr, ystr );
+          if ( error )
+            goto Next;
+        }
+
+        if ( slot->advance.x )
+          slot->advance.x += xstr;
+
+        if ( slot->advance.y )
+          slot->advance.y += ystr;
+
+        slot->metrics.width        += xstr;
+        slot->metrics.height       += ystr;
+        slot->metrics.horiAdvance  += xstr;
+        slot->metrics.vertAdvance  += ystr;
+
+        if ( slot->format == FT_GLYPH_FORMAT_BITMAP )
+          slot->bitmap_top += ystr >> 6;
+
+        error = FTDemo_Draw_Slot( handle, display, slot, &x, &y );
+
+        if ( error )
+          goto Next;
+        else if ( X_TOO_LONG( x, size, display ) )
+        {
+          x  = start_x;
+          y += step_y;
+
+          if ( Y_TOO_LONG( y, size, display ) )
+            break;
+        }
+      }
+      else
+    Next:
+        status.Fail++;
+
+      i++;
+    }
+
+    return error;
+  }
+
+
+  static FT_Error
+  Render_All( int  num_indices,
+              int  first_index )
+  {
+    int      start_x, start_y, step_y, x, y;
+    int      i;
+    FT_Size  size;
+
+
+    error = FTDemo_Get_Size( handle, &size );
+
+    if ( error )
+    {
+      /* probably a non-existent bitmap font size */
+      return error;
+    }
+
+    INIT_SIZE( size, start_x, start_y, step_y, x, y );
 
     i = first_index;
 
@@ -296,7 +444,7 @@ Next:
         status.Fail++;
       else if ( X_TOO_LONG( x, size, display ) )
       {
-        x  = start_x;
+        x = start_x;
         y += step_y;
 
         if ( Y_TOO_LONG( y, size, display ) )
@@ -314,11 +462,12 @@ Next:
   Render_Text( int  num_indices,
                int  first_index )
   {
-    int      start_x, start_y, step_x, step_y, x, y;
+    int      start_x, start_y, step_y, x, y;
     int      i;
     FT_Size  size;
 
-    const unsigned char*  p;
+    const char*  p;
+    const char*  pEnd;
 
 
     num_indices = num_indices;  /* pacify compiler */
@@ -330,24 +479,30 @@ Next:
       return error;
     }
 
-    INIT_SIZE( size, start_x, start_y, step_x, step_y, x, y );
+    INIT_SIZE( size, start_x, start_y, step_y, x, y );
 
     i = first_index;
 
-    p = Text;
+    p    = (const char*)Text;
+    pEnd = p + strlen( (const char*)Text );
 
-    while ( i > 0 && *p )
+    while ( i > 0 )
     {
-      p++;
+      utf8_next( &p, pEnd );
       i--;
     }
 
-    while ( *p && num_indices != 0 )
+    while ( num_indices != 0 )
     {
       FT_UInt  gindex;
+      int      ch;
 
 
-      gindex = FTDemo_Get_Index( handle, *p );
+      ch = utf8_next( &p, pEnd );
+      if ( ch < 0 )
+        break;
+
+      gindex = FTDemo_Get_Index( handle, ch );
 
       error = FTDemo_Draw_Index( handle, display, gindex, &x, &y );
       if ( error )
@@ -367,8 +522,6 @@ Next:
         }
       }
 
-      p++;
-
       if ( num_indices > 0 )
         num_indices -= 1;
     }
@@ -380,45 +533,44 @@ Next:
   static FT_Error
   Render_Waterfall( int  first_size )
   {
-    int         start_x, start_y, step_x, step_y, x, y;
-    int         pt_size, max_size = 100000;
-    FT_Size     size;
-    FT_Face     face;
+    int      start_x, start_y, step_y, x, y;
+    int      pt_size, max_size = 100000;
+    FT_Size  size;
+    FT_Face  face;
 
     unsigned char         text[256];
     const unsigned char*  p;
 
 
+    error = FTC_Manager_LookupFace( handle->cache_manager,
+                                    handle->scaler.face_id, &face );
+    if ( error )
     {
-      error = FTC_Manager_LookupFace( handle->cache_manager,
-                                      handle->scaler.face_id, &face );
-      if ( error )
-      {
-        /* can't access the font file. do not render anything */
-        fprintf( stderr, "can't access font file %p\n", (void*)handle->scaler.face_id );
-        return 0;
-      }
+      /* can't access the font file: do not render anything */
+      fprintf( stderr, "can't access font file %p\n",
+               (void*)handle->scaler.face_id );
+      return 0;
+    }
 
-      if ( !FT_IS_SCALABLE( face ) )
-      {
-        int  i;
+    if ( !FT_IS_SCALABLE( face ) )
+    {
+      int  i;
 
 
-        max_size = 0;
-        for ( i = 0; i < face->num_fixed_sizes; i++ )
-          if ( face->available_sizes[i].height >= max_size/64 )
-            max_size = face->available_sizes[i].height*64;
-      }
+      max_size = 0;
+      for ( i = 0; i < face->num_fixed_sizes; i++ )
+        if ( face->available_sizes[i].height >= max_size / 64 )
+          max_size = face->available_sizes[i].height * 64;
     }
 
     start_x = 4;
-    start_y = 2 * HEADER_HEIGHT;
+    start_y = 3 * HEADER_HEIGHT;
 
     for ( pt_size = first_size; pt_size < max_size; pt_size += 64 )
     {
       sprintf( (char*)text,
-                "%g: the quick brown fox jumps over the lazy dog "
-                "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789", pt_size/64.0 );
+               "%g: the quick brown fox jumps over the lazy dog"
+               " ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789", pt_size / 64.0 );
 
       p = text;
 
@@ -431,7 +583,6 @@ Next:
         continue;
       }
 
-      step_x = ( size->metrics.max_advance >> 6 ) + 4;
       step_y = ( size->metrics.height >> 6 ) + 1;
 
       x = start_x;
@@ -491,40 +642,41 @@ Next:
     grLn();
     grWriteln( "Use the following keys:" );
     grLn();
-    grWriteln( "  F1 or ?    : display this help screen" );
+    grWriteln( "  F1, ?       display this help screen" );
     grLn();
-    grWriteln( "  a          : toggle anti-aliasing" );
-    grWriteln( "  b          : toggle embedded bitmaps" );
-    grWriteln( "  c          : toggle between cache modes" );
-    grWriteln( "  f          : toggle forced auto-hinting" );
-    grWriteln( "  h          : toggle outline hinting" );
-    grWriteln( "  l          : toggle low precision rendering" );
+    grWriteln( "  a           toggle anti-aliasing" );
+    grWriteln( "  b           toggle embedded bitmaps" );
+    grWriteln( "  c           toggle between cache modes" );
+    grWriteln( "  f           toggle forced auto-hinting" );
+    grWriteln( "  h           toggle outline hinting" );
+    grWriteln( "  l           toggle low precision rendering" );
     grLn();
-    grWriteln( "  L          : cycle through LCD modes" );
-    grWriteln( "  space      : toggle rendering mode" );
-    grWriteln( "  1-5        : select rendering mode" );
+    grWriteln( "  K           cycle backwards through LCD modes" );
+    grWriteln( "  L           cycle forwards through LCD modes" );
+    grWriteln( "  backspace   cycle backwards through rendering modes" );
+    grWriteln( "  space       cycle forwards through rendering modes" );
+    grWriteln( "  1-6         select rendering mode" );
     grLn();
-    grWriteln( "  G          : show gamma ramp" );
-    grWriteln( "  g          : increase gamma by 0.1" );
-    grWriteln( "  v          : decrease gamma by 0.1" );
+    grWriteln( "  x, X        adjust horizontal emboldening" );
+    grWriteln( "  y, Y        adjust vertical emboldening" );
+    grWriteln( "  s, S        adjust slanting" );
     grLn();
-    grWriteln( "  n          : next font" );
-    grWriteln( "  p          : previous font" );
+    grWriteln( "  F           toggle custom LCD filter mode" );
+    grWriteln( "  [, ]        select custom LCD filter weight" );
+    grWriteln( "  -, +(=)     adjust selected custom LCD filter weight" );
     grLn();
-    grWriteln( "  Up         : increase pointsize by 1 unit" );
-    grWriteln( "  Down       : decrease pointsize by 1 unit" );
-    grWriteln( "  Page Up    : increase pointsize by 10 units" );
-    grWriteln( "  Page Down  : decrease pointsize by 10 units" );
+    grWriteln( "  G           show gamma ramp" );
+    grWriteln( "  g, v        adjust gamma value" );
     grLn();
-    grWriteln( "  Right      : increment index" );
-    grWriteln( "  Left       : decrement index" );
+    grWriteln( "  p, n        select previous/next font" );
     grLn();
-    grWriteln( "  F7         : decrement index by 10" );
-    grWriteln( "  F8         : increment index by 10" );
-    grWriteln( "  F9         : decrement index by 100" );
-    grWriteln( "  F10        : increment index by 100" );
-    grWriteln( "  F11        : decrement index by 1000" );
-    grWriteln( "  F12        : increment index by 1000" );
+    grWriteln( "  Up, Down    adjust pointsize by 1 unit" );
+    grWriteln( "  PgUp, PgDn  adjust pointsize by 10 units" );
+    grLn();
+    grWriteln( "  Left, Right adjust index by 1" );
+    grWriteln( "  F7, F8      adjust index by 10" );
+    grWriteln( "  F9, F10     adjust index by 100" );
+    grWriteln( "  F11, F12    adjust index by 1000" );
     grLn();
     grWriteln( "press any key to exit this help screen" );
 
@@ -542,8 +694,8 @@ Next:
     int      xside  = 10;
     int      levels = 17;
     int      gammas = 30;
-    int      x_0    = (display->bitmap->width - levels*xside)/2;
-    int      y_0    = (display->bitmap->rows  - gammas*(yside+1))/2;
+    int      x_0    = ( display->bitmap->width - levels * xside ) / 2;
+    int      y_0    = ( display->bitmap->rows - gammas * ( yside + 1 ) ) / 2;
     int      pitch  = display->bitmap->pitch;
 
 
@@ -553,18 +705,21 @@ Next:
     if ( pitch < 0 )
       pitch = -pitch;
 
-    memset( display->bitmap->buffer, 100, pitch*display->bitmap->rows );
+    memset( display->bitmap->buffer, 100, pitch * display->bitmap->rows );
 
-    grWriteCellString( display->bitmap, 0, 0, "Gamma grid", display->fore_color );
+    grWriteCellString( display->bitmap, 0, 0, "Gamma grid",
+                       display->fore_color );
 
     for ( g = 1; g <= gammas; g += 1 )
     {
-      double ggamma = g / 10.0;
-      char   temp[6];
-      int    y = y_0 + ( yside + 1 ) * ( g - 1 );
-      int    nx, ny;
+      double  ggamma = g / 10.0;
+      char    temp[6];
+      int     y = y_0 + ( yside + 1 ) * ( g - 1 );
+      int     nx, ny;
+
+      unsigned char*  line = display->bitmap->buffer +
+                             y * display->bitmap->pitch;
 
-      unsigned char*  line = display->bitmap->buffer + y * display->bitmap->pitch;
 
       if ( display->bitmap->pitch < 0 )
         line -= display->bitmap->pitch * ( display->bitmap->rows - 1 );
@@ -581,24 +736,25 @@ Next:
       {
         unsigned char*  dst = line;
 
+
         for ( nx = 0; nx < levels; nx++, dst += 3 * xside )
         {
-          double  p   = nx/(double)(levels - 1);
+          double  p   = nx / (double)( levels - 1 );
           int     gm  = (int)( 255.0 * pow( p, ggamma ) );
 
+
           memset( dst, gm, xside * 3 );
         }
       }
     }
 
-
     grRefreshSurface( display->surface );
     grListenSurface( display->surface, gr_event_key, &dummy_event );
   }
 
 
   static void
-  event_gamma_change( double delta )
+  event_gamma_change( double  delta )
   {
     status.gamma += delta;
 
@@ -617,23 +773,64 @@ Next:
 
 
   static void
-  event_size_change( int delta )
+  event_bold_change( double  xdelta,
+                     double  ydelta )
+  {
+    status.xbold_factor += xdelta;
+    status.ybold_factor += ydelta;
+
+    if ( status.xbold_factor > 0.1 )
+      status.xbold_factor = 0.1;
+    else if ( status.xbold_factor < -0.1 )
+      status.xbold_factor = -0.1;
+
+    if ( status.ybold_factor > 0.1 )
+      status.ybold_factor = 0.1;
+    else if ( status.ybold_factor < -0.1 )
+      status.ybold_factor = -0.1;
+
+    sprintf( status.header_buffer, "embolding factors changed to %.3f,%.3f",
+             status.xbold_factor, status.ybold_factor );
+
+    status.header = status.header_buffer;
+  }
+
+
+  static void
+  event_slant_change( double  delta )
+  {
+    status.slant += delta;
+
+    if ( status.slant > 1.0 )
+      status.slant = 1.0;
+    else if ( status.slant < -1.0 )
+      status.slant = -1.0;
+
+    sprintf( status.header_buffer, "slanting changed to %.3f",
+             status.slant );
+
+    status.header = status.header_buffer;
+  }
+
+
+  static void
+  event_size_change( int  delta )
   {
     status.ptsize += delta;
 
-    if ( status.ptsize < 64*1 )
-      status.ptsize = 1*64;
-    else if ( status.ptsize > MAXPTSIZE*64 )
-      status.ptsize = MAXPTSIZE*64;
+    if ( status.ptsize < 64 * 1 )
+      status.ptsize = 1 * 64;
+    else if ( status.ptsize > MAXPTSIZE * 64 )
+      status.ptsize = MAXPTSIZE * 64;
 
     FTDemo_Set_Current_Charsize( handle, status.ptsize, status.res );
   }
 
 
   static void
-  event_index_change( int delta )
+  event_index_change( int  delta )
   {
-    int num_indices = handle->current_font->num_indices;
+    int  num_indices = handle->current_font->num_indices;
 
 
     status.Num += delta;
@@ -646,7 +843,7 @@ Next:
 
 
   static void
-  event_render_mode_change( int delta )
+  event_render_mode_change( int  delta )
   {
 
     if ( delta )
@@ -665,6 +862,9 @@ Next:
     case RENDER_MODE_EMBOLDEN:
       status.header = (char *)"rendering emboldened text";
       break;
+    case RENDER_MODE_SLANTED:
+      status.header = (char *)"rendering slanted text";
+      break;
     case RENDER_MODE_STROKE:
       status.header = (char *)"rendering stroked text";
       break;
@@ -681,11 +881,11 @@ Next:
   static void
   event_font_change( int  delta )
   {
-    int      num_indices;
+    int  num_indices;
 
 
     if ( status.font_index + delta >= handle->num_fonts ||
-         status.font_index + delta < 0 )
+         status.font_index + delta < 0                  )
       return;
 
     status.font_index += delta;
@@ -706,6 +906,7 @@ Next:
   {
     int  ret = 0;
 
+
     if ( event->key >= '1' && event->key < '1' + N_RENDER_MODES )
     {
       status.render_mode = event->key - '1';
@@ -738,8 +939,8 @@ Next:
     case grKEY( 'b' ):
       handle->use_sbits = !handle->use_sbits;
       status.header     = handle->use_sbits
-                           ? (char *)"embedded bitmaps are now used when available"
-                           : (char *)"embedded bitmaps are now ignored";
+                           ? (char *)"now using embedded bitmaps (if available)"
+                           : (char *)"now ignoring embedded bitmaps";
 
       FTDemo_Update_Current_Flags( handle );
       break;
@@ -779,41 +980,76 @@ Next:
       break;
 
     case grKEY( 'L' ):
-      handle->lcd_mode = ( handle->lcd_mode + 1 ) % N_LCD_MODES;
+    case grKEY( 'K' ):
+      handle->lcd_mode = ( event->key == grKEY( 'L' ) )
+                         ? ( ( handle->lcd_mode == ( N_LCD_MODES - 1 ) )
+                             ? 0
+                             : handle->lcd_mode + 1 )
+                         : ( ( handle->lcd_mode == 0 )
+                             ? ( N_LCD_MODES - 1 )
+                             : handle->lcd_mode - 1 );
 
       switch ( handle->lcd_mode )
       {
       case LCD_MODE_AA:
-        status.header = (char *)"normal anti-aliased rendering on";
+        status.header = (char *)"use normal anti-aliased rendering";
         break;
       case LCD_MODE_LIGHT:
-        status.header = (char *)"light anti-aliased rendering on";
+        status.header = (char *)"use light anti-aliased rendering";
         break;
       case LCD_MODE_RGB:
-        status.header = (char *)"horizontal LCD-optimized rendering on (RGB)";
+        status.header = (char *)"use horizontal LCD-optimized rendering (RGB)";
         break;
       case LCD_MODE_BGR:
-        status.header = (char *)"horizontal LCD-optimized rendering on (BGR)";
+        status.header = (char *)"use horizontal LCD-optimized rendering (BGR)";
         break;
       case LCD_MODE_VRGB:
-        status.header = (char *)"vertical LCD-optimized rendering on (RGB)";
+        status.header = (char *)"use vertical LCD-optimized rendering (RGB)";
         break;
       case LCD_MODE_VBGR:
-        status.header = (char *)"vertical LCD-optimized rendering on (BGR)";
+        status.header = (char *)"use vertical LCD-optimized rendering (BGR)";
         break;
       }
 
       FTDemo_Update_Current_Flags( handle );
       break;
 
-    case grKEY( ' ' ):
+    case grKeySpace:
       event_render_mode_change( 1 );
       break;
 
+    case grKeyBackSpace:
+      event_render_mode_change( -1 );
+      break;
+
     case grKEY( 'G' ):
       event_gamma_grid();
       break;
 
+    case grKEY( 's' ):
+      event_slant_change( 0.02 );
+      break;
+
+    case grKEY( 'S' ):
+      event_slant_change( -0.02 );
+      break;
+
+    case grKEY( 'x' ):
+      event_bold_change( 0.005, 0.0 );
+      break;
+
+    case grKEY( 'X' ):
+      event_bold_change( -0.005, 0.0 );
+      break;
+
+    case grKEY( 'y' ):
+      event_bold_change( 0.0, 0.005 );
+      break;
+
+    case grKEY( 'Y' ):
+      event_bold_change( 0.0, -0.005 );
+      break;
+
     case grKEY( 'g' ):
       event_gamma_change( 0.1 );
       break;
@@ -844,8 +1080,67 @@ Next:
     case grKeyF11:   event_index_change( -1000 ); break;
     case grKeyF12:   event_index_change(  1000 ); break;
 
+    case grKEY( 'F' ):
+      FTC_Manager_RemoveFaceID( handle->cache_manager,
+                                handle->scaler.face_id );
+
+      status.use_custom_lcd_filter = !status.use_custom_lcd_filter;
+      if ( status.use_custom_lcd_filter )
+        FT_Library_SetLcdFilterWeights( handle->library,
+                                        status.filter_weights );
+      else
+        FT_Library_SetLcdFilterWeights( handle->library,
+                                        (unsigned char*)"\x10\x40\x70\x40\x10" );
+      status.header = status.use_custom_lcd_filter
+                      ? (char *)"using custom LCD filter weights"
+                      : (char *)"using default LCD filter";
+      break;
+
+    case grKEY( '[' ):
+      if ( !status.use_custom_lcd_filter )
+        break;
+
+      status.fw_index--;
+      if ( status.fw_index < 0 )
+        status.fw_index = 4;
+      break;
+
+    case grKEY( ']' ):
+      if ( !status.use_custom_lcd_filter )
+        break;
+
+      status.fw_index++;
+      if ( status.fw_index > 4 )
+        status.fw_index = 0;
+      break;
+
+    case grKEY( '-' ):
+      if ( !status.use_custom_lcd_filter )
+        break;
+
+      FTC_Manager_RemoveFaceID( handle->cache_manager,
+                                handle->scaler.face_id );
+
+      status.filter_weights[status.fw_index]--;
+      FT_Library_SetLcdFilterWeights( handle->library,
+                                      status.filter_weights );
+      break;
+
+    case grKEY( '+' ):
+    case grKEY( '=' ):
+      if ( !status.use_custom_lcd_filter )
+        break;
+
+      FTC_Manager_RemoveFaceID( handle->cache_manager,
+                                handle->scaler.face_id );
+
+      status.filter_weights[status.fw_index]++;
+      FT_Library_SetLcdFilterWeights( handle->library,
+                                      status.filter_weights );
+      break;
+
     default:
-      ;
+      break;
     }
 
     return ret;
@@ -853,7 +1148,7 @@ Next:
 
 
   static void
-  write_header( FT_Error error_code )
+  write_header( FT_Error  error_code )
   {
     FT_Face      face;
     const char*  basename;
@@ -872,31 +1167,35 @@ Next:
       switch ( error_code )
       {
       case FT_Err_Ok:
-        sprintf( status.header_buffer, "%s %s (file `%s')", face->family_name,
-                 face->style_name, basename );
+        sprintf( status.header_buffer, "%.50s %.50s (file `%.100s')",
+                 face->family_name, face->style_name, basename );
         break;
       case FT_Err_Invalid_Pixel_Size:
-        sprintf( status.header_buffer, "Invalid pixel size (file `%s')", basename );
+        sprintf( status.header_buffer, "Invalid pixel size (file `%.100s')",
+                 basename );
         break;
       case FT_Err_Invalid_PPem:
-        sprintf( status.header_buffer, "Invalid ppem value (file `%s')", basename );
+        sprintf( status.header_buffer, "Invalid ppem value (file `%.100s')",
+                 basename );
         break;
       default:
-        sprintf( status.header_buffer, "File `%s': error 0x%04x", basename,
-                 (FT_UShort)error_code );
+        sprintf( status.header_buffer, "File `%.100s': error 0x%04x",
+                 basename, (FT_UShort)error_code );
         break;
       }
 
       status.header = status.header_buffer;
     }
 
-    grWriteCellString( display->bitmap, 0, 0, status.header, display->fore_color );
+    grWriteCellString( display->bitmap, 0, 0,
+                       status.header, display->fore_color );
 
-    format = ( status.encoding != FT_ENCODING_NONE )
+    format = status.encoding != FT_ENCODING_NONE
              ? "at %g points, first char code = 0x%x"
              : "at %g points, first glyph index = %d";
 
-    snprintf( status.header_buffer, 256, format, status.ptsize/64.0, status.Num );
+    snprintf( status.header_buffer, 256, format,
+              status.ptsize / 64.0, status.Num );
 
     if ( FT_HAS_GLYPH_NAMES( face ) )
     {
@@ -918,14 +1217,32 @@ Next:
           gindex = FTDemo_Get_Index( handle, status.Num );
 
         strcpy( p, format );
-        if ( FT_Get_Glyph_Name( face, gindex, p + format_len, size - format_len ) )
+        if ( FT_Get_Glyph_Name( face, gindex,
+                                p + format_len, size - format_len ) )
           *p = '\0';
       }
     }
 
     status.header = status.header_buffer;
-    grWriteCellString( display->bitmap, 0, HEADER_HEIGHT, status.header_buffer,
-                       display->fore_color );
+    grWriteCellString( display->bitmap, 0, HEADER_HEIGHT,
+                       status.header_buffer, display->fore_color );
+
+    if ( status.use_custom_lcd_filter )
+    {
+      int             fwi = status.fw_index;
+      unsigned char  *fw  = status.filter_weights;
+
+
+      sprintf( status.header_buffer,
+               "%s0x%02X%s%s0x%02X%s%s0x%02X%s%s0x%02X%s%s0x%02X%s",
+               fwi == 0 ? "[" : " ", fw[0], fwi == 0 ? "]" : " ",
+               fwi == 1 ? "[" : " ", fw[1], fwi == 1 ? "]" : " ",
+               fwi == 2 ? "[" : " ", fw[2], fwi == 2 ? "]" : " ",
+               fwi == 3 ? "[" : " ", fw[3], fwi == 3 ? "]" : " ",
+               fwi == 4 ? "[" : " ", fw[4], fwi == 4 ? "]" : " " );
+      grWriteCellString( display->bitmap, 0, 2 * HEADER_HEIGHT,
+                         status.header_buffer, display->fore_color );
+    }
 
     grRefreshSurface( display->surface );
   }
@@ -934,32 +1251,37 @@ Next:
   static void
   usage( char*  execname )
   {
-    fprintf( stderr,  "\n" );
-    fprintf( stderr,  "ftview: simple glyph viewer -- part of the FreeType project\n" );
-    fprintf( stderr,  "-----------------------------------------------------------\n" );
-    fprintf( stderr,  "\n" );
-    fprintf( stderr,  "Usage: %s [options] pt font ...\n",
-             execname );
-    fprintf( stderr,  "\n" );
-    fprintf( stderr,  "  pt        The point size for the given resolution.\n" );
-    fprintf( stderr,  "            If resolution = 72 dpi, this directly gives the\n" );
-    fprintf( stderr,  "            ppem value (pixels per EM).\n" );
-    fprintf( stderr,  "  font      The font file(s) to display; if no extension is given,\n" );
-    fprintf( stderr,  "            ftview tries to append one of the following suffixes:\n" );
-    fprintf( stderr,  "            `.ttf', `.ttc', `.otf', `.pfa', `.pfb'.\n" );
-    fprintf( stderr,  "            For `.pfa' and `.pfb' files, ftview also tries to\n" );
-    fprintf( stderr,  "            attach the corresponding metrics file (with extension\n" );
-    fprintf( stderr,  "            `.afm' or `.pfm').\n" );
-    fprintf( stderr,  "\n" );
-    fprintf( stderr,  "  -r R      Use resolution R dpi (default: 72 dpi).\n" );
-    fprintf( stderr,  "  -f index  Specify first index to display.\n" );
-    fprintf( stderr,  "  -e enc    Specify encoding tag (default: no encoding).\n" );
-    fprintf( stderr,  "  -D        Dump cache usage statistics.\n" );
-    fprintf( stderr,  "  -m text   Use `text' for rendering.\n" );
-    fprintf( stderr,  "  -l nn     Change rendering mode (0 <= nn <= %d).\n",
-             N_LCD_MODES );
-    fprintf( stderr,  "  -p        Preload file in memory to simulate memory-mapping.\n" );
-    fprintf( stderr,  "\n" );
+    fprintf( stderr,
+      "\n"
+      "ftview: simple glyph viewer -- part of the FreeType project\n"
+      "-----------------------------------------------------------\n"
+      "\n" );
+    fprintf( stderr,
+      "Usage: %s [options] pt font ...\n"
+      "\n", execname );
+    fprintf( stderr,
+      "  pt        The point size for the given resolution.\n"
+      "            If resolution is 72dpi, this directly gives the\n"
+      "            ppem value (pixels per EM).\n" );
+    fprintf( stderr,
+      "  font      The font file(s) to display.\n"
+      "            For Type 1 font files, ftview also tries to attach\n"
+      "            the corresponding metrics file (with extension\n"
+      "            `.afm' or `.pfm').\n"
+      "\n" );
+    fprintf( stderr,
+      "  -r R      Use resolution R dpi (default: 72dpi).\n"
+      "  -f index  Specify first index to display (default: 0).\n"
+      "  -e enc    Specify encoding tag (default: no encoding).\n"
+      "            Common values: `unic' (Unicode), `symb' (symbol),\n"
+      "            `ADOB' (Adobe standard), `ADBC' (Adobe custom).\n"
+      "  -D        Dump cache usage statistics.\n"
+      "  -m text   Use `text' for rendering.\n" );
+    fprintf( stderr,
+      "  -l mode   Change rendering mode (0 <= mode <= %d).\n", N_LCD_MODES );
+    fprintf( stderr,
+      "  -p        Preload file in memory to simulate memory-mapping.\n"
+      "\n" );
 
     exit( 1 );
   }
@@ -1043,9 +1365,9 @@ Next:
     if ( *argc <= 1 )
       usage( execname );
 
-    status.ptsize = atof( *argv[0] ) * 64.0;
+    status.ptsize = (int)( atof( *argv[0] ) * 64.0 );
     if ( status.ptsize == 0 )
-      status.ptsize = 64*10;
+      status.ptsize = 64 * 10;
 
     (*argc)--;
     (*argv)++;
@@ -1064,15 +1386,15 @@ Next:
 #if FREETYPE_MAJOR == 2 && FREETYPE_MINOR == 0 && FREETYPE_PATCH <= 8
     if ( status.debug )
     {
-#  ifdef FT_DEBUG_LEVEL_TRACE
+#ifdef FT_DEBUG_LEVEL_TRACE
       FT_SetTraceLevel( trace_any, (FT_Byte)status.trace_level );
-#  else
+#else
       status.trace_level = 0;
-#  endif
+#endif
     }
 #elif 0
-       /* "setenv/putenv" is not ANSI and I don't want to mess */
-       /* with this portability issue right now..              */
+       /* `setenv' and `putenv' is not ANSI and I don't want to mess */
+       /* with this portability issue right now...                   */
     if ( status.debug )
     {
       char  temp[32];
@@ -1102,7 +1424,8 @@ Next:
 
     memset( display->fore_color.chroma, 0, 4 );
     memset( display->back_color.chroma, 0xff, 4 );
-    grSetTitle( display->surface, "FreeType Glyph Viewer - press F1 for help" );
+    grSetTitle( display->surface,
+                "FreeType Glyph Viewer - press F1 for help" );
 
     status.Fail = 0;
 
@@ -1120,15 +1443,23 @@ Next:
       switch ( status.render_mode )
       {
       case RENDER_MODE_ALL:
-        error = Render_All( handle->current_font->num_indices, status.Num );
+        error = Render_All( handle->current_font->num_indices,
+                            status.Num );
         break;
 
       case RENDER_MODE_EMBOLDEN:
-        error = Render_Embolden( handle->current_font->num_indices, status.Num );
+        error = Render_Embolden( handle->current_font->num_indices,
+                                 status.Num );
+        break;
+
+      case RENDER_MODE_SLANTED:
+        error = Render_Slanted( handle->current_font->num_indices,
+                                status.Num );
         break;
 
       case RENDER_MODE_STROKE:
-        error = Render_Stroke( handle->current_font->num_indices, status.Num );
+        error = Render_Stroke( handle->current_font->num_indices,
+                               status.Num );
         break;
 
       case RENDER_MODE_TEXT:
